import copy
from open_dread_rando.text_patches import apply_text_patches, patch_text
from open_dread_rando.patcher_editor import PatcherEditor
from mercury_engine_data_structures.formats.dread_types import CTriggerComponent_EEvent


# when playing with artifacts, ensure that sufficient artifact items are shuffled
# artifact item IDs follow the pattern "ITEM_RANDO_ARTIFACT_N" where N is an integer in [1, required_artifacts]
def apply_objective_patches(editor: PatcherEditor, configuration: dict):
    if configuration["objective"]["required_artifacts"] == 0:
        return
    
    patch_text(
        editor,
        "RANDO_ARTIFACTS_ALL",
        r"All Metroid DNA acquired.|{c3}Raven Beak{c0} awaits you in {c5}Itorash{c0}."
    )

    access_point = copy.deepcopy(editor.resolve_actor_reference({
        "scenario": "s080_shipyard",
        "actor": "accesspoint_000"
    }))
    ap_platform = copy.deepcopy(editor.resolve_actor_reference({
        "scenario": "s080_shipyard",
        "actor": "weightactivatedplatform_access_000"
    }))
    ap_trigger = copy.deepcopy(editor.resolve_actor_reference({
        "scenario": "s080_shipyard",
        "actor": "AP_10"
    }))

    origin = access_point.vPos
    new_origin = (-2232.221 - origin[0], -3500.0 - origin[1], 0.0)
    itorash = editor.get_scenario("s090_skybase")

    # add actors to Itorash and update their position
    for actor in [access_point, ap_platform, ap_trigger]:
        itorash.actors_for_layer('default')[actor.sName] = actor
        itorash.add_actor_to_group('eg_collision_camera_001_Default', actor.sName)
        actor.vPos = [c + offset for c, offset in zip(actor.vPos, new_origin)]
    
    for charclass in [
        "actors/props/weightactivatedplatform_access",
        "actors/props/accesspoint",
        "actors/logic/trigger",
    ]:
        for asset in editor.get_asset_names_in_folder(charclass):
            editor.ensure_present_in_scenario("s090_skybase", asset)
    
    trigger = ap_trigger.pComponents.TRIGGER
    on_exit = copy.deepcopy(trigger.lstActivationConditions[0])
    on_exit.eEvent = CTriggerComponent_EEvent.OnExit
    on_exit.vLogicActions[0].sCallback = "CurrentScenario.OnExit_AP_10"
    trigger.lstActivationConditions.append(on_exit)

    # increase the height of the trigger to prevent jumping over it
    segments = ap_trigger.pComponents.LOGICSHAPE.pLogicShape.oPolyCollection.vPolys[0].oSegmentData
    for i in range(2):
        segments[i].vPos[1] += 1000

    usable = access_point.pComponents.USABLE

    usable.vDoorsToChange = [itorash.link_for_actor('doorpowerpower_000')]
    usable.wpThermalDevice = ""

    hints = {
        f"DIAG_ADAM_SHIP_2_PAGE_{i+1}": hint
        for i, hint in enumerate(configuration['objective']['hints'])
    }
    apply_text_patches(editor, hints)
    
    mlogs = {
        f"MLOG_ITEM_RANDO_ARTIFACT_{i+1}": f"Acquired {r'{c1}'}Metroid DNA {i+1}{r'{c0}'}"
        for i in range(12)
    }
    apply_text_patches(editor, mlogs)

    usable.tCaptionList = {
        "DIAG_ADAM_SHIP_2": list(hints) or ["DIAG_ADAM_SHIP_2_PAGE_1"]
    }

    quiet_robe_usable = editor.resolve_actor_reference({
        "scenario": "s030_baselab",
        "layer": "default",
        "actor": "accesspoint_001"
    })
    quiet_robe_usable.pComponents.USABLE.tCaptionList = {
        "DIAG_ADAM_LAB_2": list(hints) or ["DIAG_ADAM_LAB_2_PAGE_1"]
    }



import shutil
import typing
from pathlib import Path

from construct import Container
from mercury_engine_data_structures.file_tree_editor import FileTreeEditor
from mercury_engine_data_structures.formats import BaseResource, Brfld, Brsa, ALL_FORMATS, Bmmap

T = typing.TypeVar("T")


def path_for_level(level_name: str) -> str:
    return f"maps/levels/c10_samus/{level_name}/{level_name}"


def extension_for_type(type_hint: typing.Type[T]) -> str:
    return next(ext for ext, t in ALL_FORMATS.items() if t == type_hint).lower()


class PatcherEditor(FileTreeEditor):
    memory_files: dict[str, BaseResource]

    def __init__(self, root: Path):
        super().__init__(root)
        self.memory_files = {}

    def get_file(self, path: str, type_hint: typing.Type[T] = BaseResource) -> T:
        if path not in self.memory_files:
            self.memory_files[path] = self.get_parsed_asset(path, type_hint=type_hint)
        return self.memory_files[path]

    def get_scenario_file(self, name: str, type_hint: typing.Type[T]) -> T:
        path = f"{path_for_level(name)}.{extension_for_type(type_hint)}"
        return self.get_file(path, type_hint)

    def get_scenario(self, name: str) -> Brfld:
        return self.get_scenario_file(name, Brfld)

    def get_subarea_manager(self, name: str) -> Brsa:
        return self.get_scenario_file(name, Brsa)

    def get_scenario_map(self, name: str) -> Bmmap:
        return self.get_scenario_file(name, Bmmap)

    def get_level_pkgs(self, name: str) -> set[str]:
        return set(self.find_pkgs(path_for_level(name) + ".brfld"))
    
    def ensure_present_in_scenario(self, scenario: str, asset):
        for pkg in self.get_level_pkgs(scenario):
            self.ensure_present(pkg, asset)

    def resolve_actor_reference(self, ref: dict) -> Container:
        scenario = self.get_scenario(ref["scenario"])
        layer = ref.get("layer", "default")
        return scenario.actors_for_layer(layer)[ref["actor"]]

    def flush_modified_assets(self):
        for name, resource in self.memory_files.items():
            self.replace_asset(name, resource)
        self.memory_files = {}

    def add_new_asset(self, name: str, new_data: typing.Union[bytes, BaseResource], in_pkgs: typing.Iterable[str]):
        super().add_new_asset(name, new_data, in_pkgs)
        # Hack for textures' weird folder layout
        if name.startswith("textures/") and isinstance(new_data, bytes):
            self._toc.add_file(name[9:], len(new_data))

    def save_modified_saves_to(self, debug_path: Path):
        shutil.rmtree(debug_path, ignore_errors=True)
        for asset_id, asset in self._modified_resources.items():
            if asset is not None:
                path = debug_path.joinpath(self._name_for_asset_id[asset_id])
                path.parent.mkdir(parents=True, exist_ok=True)
                path.write_bytes(asset)

    def remove_entity(self, reference: dict, map_category: typing.Optional[str]):
        scenario = self.get_scenario(reference["scenario"])
        layer = reference.get("layer", "default")
        actor_name = reference["actor"]

        for group_name in scenario.all_actor_groups():
            scenario.remove_actor_from_group(group_name, actor_name, layer)

        scenario.actors_for_layer(layer).pop(actor_name)
        if map_category is not None:
            self.get_scenario_map(reference["scenario"]).raw.Root[map_category].pop(actor_name)
    
    def get_asset_names_in_folder(self, folder: str) -> typing.Iterator[str]:
        yield from (name for name in self._name_for_asset_id.values() if name.startswith(folder))



from mercury_engine_data_structures.formats import Bmsad

from open_dread_rando.patcher_editor import PatcherEditor

from open_dread_rando.environmental_damage_sources import ALL_DAMAGE_ROOM_ACTORS
from open_dread_rando.text_patches import patch_text

from math import ceil

_HANUBIA_SHORTCUT_GRAPPLE_BLOCKS = [
    {
        "scenario": "s080_shipyard",
        "layer": "default",
        "actor": "grapplepulloff1x2_000"
    },
    {
        "scenario": "s080_shipyard",
        "layer": "default",
        "actor": "grapplepulloff1x2"
    }
]


def apply_game_patches(editor: PatcherEditor, configuration: dict):
    if configuration.get("consistent_raven_beak_damage_table", True):
        _consistent_raven_beak_damage_table(editor)

    _remove_grapple_blocks(editor, configuration)
    
    if configuration["linear_damage_runs"]:
        _linear_damage_runs(editor, configuration)
    
    if configuration["warp_to_start"]:
        _warp_to_start(editor)
        
            

def _consistent_raven_beak_damage_table(editor: PatcherEditor):
    rb_bmsad = editor.get_file("actors/characters/chozocommander/charclasses/chozocommander.bmsad", Bmsad)

    life_component = rb_bmsad.raw.property.components.LIFE
    ai_component = rb_bmsad.raw.property.components.AI
    factors = [
        life_component.fields.fields.oDamageSourceFactor,
        ai_component.fields.fields.oDamageSourceFactorShortShootingGrab,
        ai_component.fields.fields.oDamageSourceFactorLongShootingGrab,
    ]

    for factor in factors:
        for beam in ["fPowerBeamFactor", "fWideBeamFactor", "fPlasmaBeamFactor"]:
            factor[beam] = factor.fWaveBeamFactor
        for beam in ["fChargePowerBeamFactor", "fChargeWideBeamFactor", "fChargePlasmaBeamFactor"]:
            factor[beam] = factor.fChargeWaveBeamFactor
        for beam in ["fMeleeChargePowerBeamFactor", "fMeleeChargeWideBeamFactor", "fMeleeChargePlasmaBeamFactor"]:
            factor[beam] = factor.fMeleeChargeWaveBeamFactor
        for missile in ["fMissileFactor", "fSuperMissileFactor"]:
            factor[missile] = factor.fIceMissileFactor

def _remove_grapple_blocks(editor: PatcherEditor, configuration: dict):
    if configuration["remove_grapple_blocks_hanubia_shortcut"]:
        for reference in _HANUBIA_SHORTCUT_GRAPPLE_BLOCKS:
            editor.remove_entity(reference, "mapProps")

    if configuration["remove_grapple_block_path_to_itorash"]:
        editor.remove_entity(
            {
                "scenario": "s080_shipyard",
                "layer": "default",
                "actor": "grapplepulloff1x2_001"
            },
            "mapProps"
        )

def _linear_damage_runs(editor: PatcherEditor, configuration: dict):
    if configuration["linear_dps"] <= 0:    #This case circumvents a potential divide-by-zero from the tick length calculation, as well as a negative tick length.
    
        damage = 0          #Damage value of 0 to bypass divide-by-zero later
        tick = 0.5          #Tick length of 0.5 is default for damage rooms, lava uses 1 second, but this should not matter.
    elif configuration["linear_dps"] > 0:
        
        #Damage per tick (or 'fDamagePerTime') must be a whole number, this calculation finds a value of damage that will be multiplied by the tick length to equal the total DPS, such that the tick length is between 0.2s and 0.4s
        damage = ceil(float(configuration["linear_dps"])/5)
    
        #Tick length (or 'fInBetweenDamageTime') is found by dividing the damage value by the DPS input, and a value of 1.5/60 is subtracted from the result.
        tick = damage/float(configuration["linear_dps"]) - (1.5/60) #1/60 is meant to account for the frame that damage is applied to Samus, which does not count toward the timer, but 1.5/60 worked more accurately in testing for an unknown reason.
    
    
    for reference in ALL_DAMAGE_ROOM_ACTORS:
    
        actor = editor.resolve_actor_reference(reference)
        
        if actor.oActorDefLink == "actordef:actors/props/env_frozen_gen_001/charclasses/env_frozen_gen_001.bmsad":
            config_name = 'oFreezeConfig'
        elif actor.oActorDefLink == "actordef:actors/props/env_heat_gen_001/charclasses/env_heat_gen_001.bmsad":
            config_name = 'oHeatConfig'
        elif actor.oActorDefLink == "actordef:actors/props/lavazone/charclasses/lavazone.bmsad":
            config_name = 'oConfig'
        else:
            raise ValueError(f"{reference} does not have a valid actorDef for environmental damage")

        damage_config = actor.pComponents.ACTIVATABLE[config_name]
        damage_config.fDamagePerTime = damage
        damage_config.fInBetweenDamageTime = tick
        damage_config.fDamageIncreaseAmount = 0     #Setting this value to 0 removes the damage ramp that exists in vanilla for damage rooms.
        damage_config.fMaxDamage = 1000             #1000 Max damage is an arbitrary limit, also set as the maximum in Randovania. There is no reason this could not be larger if needed.

def _warp_to_start(editor: PatcherEditor):
    text = "Save your progress?|Hold \u1806 and \u1807 while selecting {c6}Cancel{c0}|to warp to the starting location."
    patch_text(editor, "GUI_SAVESTATION_QUESTION", text)



_ARTARIA_HEAT_ROOM_ACTORS = [
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_001"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_002"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_003"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_004"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_CR_001"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_CR_002"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_CR_003"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_CR_004"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_CR_005"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_heat_gen_001_CR_END"
    }
]

_ARTARIA_COLD_ROOM_ACTORS = [
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_frozen_gen_001_000"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_frozen_gen_001_001"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "env_frozen_gen_001_002"
    },
]

_ARTARIA_LAVA_ACTORS = [
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "lavazone_001"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "lavazone_002"
    },
    {
        "scenario": "s010_cave",
        "layer": "default",
        "actor": "lavazone_003"
    },
]

_CATARIS_HEAT_ROOM_ACTORS = [
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "env_heat_gen_001"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "env_heat_gen_002"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "env_heat_gen_003"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "env_heat_gen_004"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "env_heat_gen_005"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "env_heat_gen_006"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "env_heat_gen_007"
    },
]

_CATARIS_LAVA_ACTORS = [
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_000"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_001"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_002"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_003"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_004"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_005"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_006"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_007"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_008"
    },
    {

        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_009"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_010"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_011"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_012"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_013"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_014"
    },
    {
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "lavazone_015"
    },
]

_DAIRON_HEAT_ROOM_ACTORS = [
    {
        "scenario": "s030_baselab",
        "layer": "default",
        "actor": "env_heat_gen_001"
    }
]

_DAIRON_COLD_ROOM_ACTORS = [
    {
        "scenario": "s030_baselab",
        "layer": "default",
        "actor": "env_frozen_gen_001"
    }
]

_DAIRON_LAVA_ACTORS = [
    {
        "scenario": "s030_baselab",
        "layer": "default",
        "actor": "lavazone_000"
    },
    {
        "scenario": "s030_baselab",
        "layer": "default",
        "actor": "lavazone_001"
    },
    {
        "scenario": "s030_baselab",
        "layer": "default",
        "actor": "lavazone_002"
    }
]

_FERENIA_COLD_ROOM_ACTORS = [
    {
        "scenario": "s070_basesanc",
        "layer": "default",
        "actor": "env_frozen_gen_000"
    },
    {
        "scenario": "s070_basesanc",
        "layer": "default",
        "actor": "env_frozen_gen_001"
    }
]

ALL_DAMAGE_ROOM_ACTORS = _ARTARIA_HEAT_ROOM_ACTORS + _ARTARIA_COLD_ROOM_ACTORS + _ARTARIA_LAVA_ACTORS + _CATARIS_HEAT_ROOM_ACTORS + _CATARIS_LAVA_ACTORS + _DAIRON_HEAT_ROOM_ACTORS + _DAIRON_COLD_ROOM_ACTORS + _DAIRON_LAVA_ACTORS + _FERENIA_COLD_ROOM_ACTORS



import copy

import construct
from mercury_engine_data_structures.formats.dread_types import CActor, CTriggerComponent_EEvent

from open_dread_rando import door_patcher
from open_dread_rando.logger import LOG
from open_dread_rando.patcher_editor import PatcherEditor


def flip_icon_id(icon_id: str) -> str:
    if icon_id.endswith("R"):
        return icon_id[:-1] + "L"
    elif icon_id.endswith("L"):
        return icon_id[:-1] + "R"
    raise ValueError(f"Unable to flip icon {icon_id}")


def apply_one_sided_door_fixes(editor: PatcherEditor):
    all_scenarios = [
        "s010_cave",
        "s020_magma",
        "s030_baselab",
        "s040_aqua",
        "s050_forest",
        "s060_quarantine",
        "s070_basesanc",
        "s080_shipyard",
    ]

    for scenario_name in all_scenarios:
        scenario = editor.get_scenario(scenario_name)
        bmmap = editor.get_scenario_map(scenario_name)
        map_blockages = bmmap.raw.Root.mapBlockages

        for layer_name, actor_name, actor in list(scenario.all_actors()):
            if not door_patcher.is_door(actor):
                continue

            if actor.oActorDefLink != "actordef:actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad":
                continue

            life_comp = actor.pComponents.LIFE
            left = scenario.follow_link(life_comp.wpLeftDoorShieldEntity)
            right = scenario.follow_link(life_comp.wpRightDoorShieldEntity)

            if left is None and right is None:
                continue

            elif left is None:
                other = right
                direction = "wpLeftDoorShieldEntity"

            elif right is None:
                other = left
                direction = "wpRightDoorShieldEntity"
            else:
                continue

            if "db_hdoor" in other.oActorDefLink:
                continue

            LOG.debug("%s/%s/%s: copy %s into %s", scenario_name, layer_name, actor_name, other.sName, direction)
            mirrored = copy.deepcopy(other)
            mirrored.sName += "_mirrored"
            mirrored.vAng = [other.vAng[0], -other.vAng[1], other.vAng[2]]
            scenario.actors_for_layer(layer_name)[mirrored.sName] = mirrored

            mirrored_map = copy.deepcopy(map_blockages[other.sName])
            mirrored_map["sIconId"] = flip_icon_id(mirrored_map["sIconId"])
            delta = 450 if mirrored_map["sIconId"].endswith("R") else -450
            mirrored_map["oBox"]["Min"][0] += delta
            mirrored_map["oBox"]["Max"][0] += delta
            map_blockages[mirrored.sName] = mirrored_map

            # Add a reference to the other shield to the main actor
            life_comp[direction] = f"Root:pScenario:rEntitiesLayer:dctSublayers:{layer_name}:dctActors:{mirrored.sName}"

            for group_name in scenario.all_actor_groups():
                if any(scenario.is_actor_in_group(group_name, x, layer_name) for x in [actor_name, other.sName]):
                    for name in [actor_name, mirrored.sName, other.sName]:
                        scenario.add_actor_to_group(group_name, name, layer_name)


PROBLEM_LAYERS = {
    "PostXRelease": {
        "s010_cave": [
            "collision_camera_026",  # Chain reaction
            "collision_camera_020",  # Corpius arena
            "collision_camera_073",  # Corpius entrance
        ],
        "s020_magma": [
            "collision_camera_063",  # Kraid arena
        ],
        "s040_aqua": [
            "collision_camera_007",  # Below Drogyga
            "collision_camera_028",  # Drogyga arena
        ],
        "s070_basesanc": [
            "collision_camera_005",  # Quiet Robe room
        ]
    },
    "Cooldown": {
        "s020_magma": [
            "collision_camera_004", # Z-57 Access
        ]
    },
    "PostEmmy": {
        "s070_basesanc": [
            "collision_camera_040", # Purple Emmi Introduction
        ]
    }
}


def remove_problematic_x_layers(editor: PatcherEditor):
    # these X layers have priority, so they will set some rooms to their post-state
    # even if they've never been entered. in these particular problem rooms, this
    # can cause softlocks (e.g. phantom cloak on golzuna, main PBs on corpius)
    for setup, levels in PROBLEM_LAYERS.items():
        for level, layers in levels.items():
            manager = editor.get_subarea_manager(level)
            configs = manager.get_subarea_setup(setup).vSubareaConfigs
            manager.get_subarea_setup(setup).vSubareaConfigs = [
                config for config in configs if config.sId not in layers
            ]


def _apply_boss_cutscene_fixes(editor: PatcherEditor, cutscene_ref: dict, callback: str):
    cutscene_player = editor.resolve_actor_reference(cutscene_ref)

    # Remove the checkpoint save
    cutscene_player.pComponents.CUTSCENE.vctOnAfterCutsceneEndsLA.pop()

    # Add the custom call when the boss dies
    if callback:
        cutscene_player.pComponents.CUTSCENE.vctOnAfterCutsceneEndsLA.append({
            "@type": "CLuaCallsLogicAction",
            "sCallbackEntityName": "",
            "sCallback": callback,
            "bCallbackEntity": False,
            "bCallbackPersistent": False,
        })

def apply_kraid_fixes(editor: PatcherEditor):
    magma = editor.get_scenario("s020_magma")

    _apply_boss_cutscene_fixes(editor, {
        "scenario": "s020_magma",
        "layer": "cutscenes",
        "actor": "cutsceneplayer_61"
    }, "CurrentScenario.OnKraidDeath_CUSTOM")

    # Remove the checkpoint after killing Kraid
    # (not sure if this is necessary)
    kraid_spawn = magma.follow_link(
        "Root:pScenario:rEntitiesLayer:dctSublayers:Boss:dctActors:SP_Kraid"
    )
    kraid_file = kraid_spawn.pComponents.SPAWNPOINT.voActorBlueprint[0]
    kraid = CActor.parse(kraid_file.InnerValue)
    kraid.pComponents.AI.wpDeadCheckpointStartPoint = "{EMPTY}"
    kraid_file.InnerValue = CActor.build(kraid)

def apply_drogyga_fixes(editor: PatcherEditor):
    _apply_boss_cutscene_fixes(editor, {
        "scenario": "s040_aqua",
        "layer": "cutscenes",
        "actor": "cutsceneplayer_65"
    }, "CurrentScenario.OnHydrogigaDead_CUSTOM")

    # remove the trigger that deletes drogyga until after beating drogyga
    aqua = editor.get_scenario("s040_aqua")
    aqua.remove_actor_from_group("eg_collision_camera_007_Default", "TG_WaterPoolAfterHydrogiga", "Boss")

def activate_emmi_zones(editor: PatcherEditor):
    # Remove the cutscene that plays when you enter the emmi zone for the first time
    # This causes the border of the zone to not be revealed, but it should be possible to do that in another way
    editor.remove_entity(
        {
            "scenario": "s010_cave",
            "layer": "Cutscenes",
            "actor": "cutscenetrigger_36"
        },
        None,
    )

    # Remove the cutscene that introduces the speed emmi
    # This causes the border of the zone to not be revealed, but it should be possible to do that in another way
    editor.remove_entity(
        {
            "scenario": "s030_baselab",
            "layer": "cutscenes",
            "actor": "cutscenetrigger_39"
        },
        None,
    )


def fix_backdoor_white_cu(editor: PatcherEditor):
    new_door = construct.Container({
        "@type": "CEntity",
        "sName": "DreadRando_CUDoor",
        "oActorDefLink": "actordef:actors/props/doorclosedpower/charclasses/doorclosedpower.bmsad",
        "vPos": [
            8300.000,
            -4700.000,
            0.0
        ],
        "vAng": [
            0.0,
            0.0,
            0.0
        ],
        "pComponents": {
            "LIFE": {
                "@type": "CDoorLifeComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "fMaxDistanceOpened": 400.0,
                "wpLeftDoorShieldEntity": "{EMPTY}",
                "wpRightDoorShieldEntity": "{EMPTY}",
                "fMinTimeOpened": 3.0,
                "bStayOpen": True,
                "bStartOpened": False,
                "bOnBlackOutOpened": False,
                "bDoorIsWet": False,
                "bFrozenDuringColdown": True,
                "iAreaLeft": 0,
                "iAreaRight": 0,
                "aVignettes": []
            },
            "AUDIO": {
                "@type": "CAudioComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "FX": {
                "@type": "CFXComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "fSelectedHighRadius": 250.0,
                "fSelectedLowRadius": 350.0
            },
            "COLLISION": {
                "@type": "CCollisionComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "TIMELINECOMPONENT": {
                "@type": "CTimelineComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "NAVMESHITEM": {
                "@type": "CNavMeshItemComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "tForbiddenEdgesSpawnPoints": []
            },
            "ANIMATION": {
                "@type": "CAnimationComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "MODELUPDATER": {
                "@type": "CModelUpdaterComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "sDefaultModelPath": ""
            }
        },
        "bEnabled": True
    })

    cave = editor.get_scenario("s010_cave")
    cave.actors_for_layer("default")[new_door["sName"]] = new_door

    for group in ["eg_collision_camera_018_Default", "eg_collision_camera_090_Default",
                  "eg_collision_camera_049_Default", "eg_collision_camera_090_PostXRelease",
                  "eg_collision_camera_049_PostXRelease"]:
        cave.add_actor_to_group(group, new_door["sName"])


def patch_corpius_checkpoints(editor: PatcherEditor):
    cave = editor.get_scenario("s010_cave")

    # Remove CSaveGameLogicAction for SP_Scorpius_scorpius_dead on the cutscene after killing Corpius
    cutscene = cave.follow_link("Root:pScenario:rEntitiesLayer:dctSublayers:Cutscenes:dctActors:cutsceneplayer_57")
    cutscene.pComponents.CUTSCENE.vctOnAfterCutsceneEndsLA.pop(0)

    # Fix checkpoint before Corpius fight not working if the X were released
    cave.add_actor_to_group("eg_collision_camera_072_PostXRelease", "SP_Checkpoint_Scorpius")


def apply_experiment_fixes(editor: PatcherEditor):
    magma = editor.get_scenario("s020_magma")

    _apply_boss_cutscene_fixes(editor, {
        "scenario": "s020_magma",
        "layer": "cutscenes",
        "actor": "cutsceneplayer_81"
    }, "")

    new_triggers = {
        "TriggerEnableCooldown": (5050.000, -5346.150, 0.000),
        "TriggerDisableCooldown": (5000.000, -7350.000, 0.000),
    }

    # create triggers to update the cooldown status appropriately
    for name, pos in new_triggers.items():
        ap_trigger = copy.deepcopy(editor.resolve_actor_reference({
            "scenario": "s020_magma",
            "layer": "default",
            "actor": "AP_03"
        }))

        ap_trigger.sName = name
        ap_trigger.vPos = pos

        ap_trigger.pComponents.TRIGGER.lstActivationConditions[0].vLogicActions[0].sCallback = f"CurrentScenario.OnEnter_{name}"

        magma.actors_for_layer('default')[name] = ap_trigger
        magma.add_actor_to_group("eg_collision_camera_004_PostXRelease", name)
    
    # make thermal doors always closed during the fight
    for name in ["trap_thermal_horizontal_000", "trap_thermal_horizontal_005"]:
        magma.remove_actor_from_group("eg_collision_camera_009_Cooldown", name)
        
        trap = copy.deepcopy(editor.resolve_actor_reference({
            "scenario": "s020_magma",
            "layer": "default",
            "actor": name
        }))

        trap.sName = f"{name}_EXPERIMENT"
        magma.actors_for_layer('default')[trap.sName] = trap
        magma.add_actor_to_group('eg_collision_camera_009_Cooldown', trap.sName)
    
    # disable closing the thermal door permanently after experiment
    editor.remove_entity({
        "scenario": "s020_magma",
        "layer": "default",
        "actor": "trap_thermal_horizontal_POSTCOOL"
    }, "mapDoors")


def apply_static_fixes(editor: PatcherEditor):
    remove_problematic_x_layers(editor)
    activate_emmi_zones(editor)
    apply_one_sided_door_fixes(editor)
    apply_kraid_fixes(editor)
    fix_backdoor_white_cu(editor)
    patch_corpius_checkpoints(editor)
    apply_experiment_fixes(editor)
    apply_drogyga_fixes(editor)



import argparse
import json
import logging
import logging.config
import time
from pathlib import Path

from open_dread_rando import dread_patcher


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input-path", required=True, type=Path,
                        help="Path to where the extracted Metroid Dread romfs to randomize can be found.")
    parser.add_argument("--output-path", required=True, type=Path,
                        help="Path to where the modified files will be written to.")
    parser.add_argument("--input-json", type=Path,
                        help="Path to the configuration json. If missing, it's read from standard input")
    return parser


def setup_logging():
    handlers = {
        'default': {
            'level': 'DEBUG',
            'formatter': 'default',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',  # Default is stderr
        },
    }
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s',
            }
        },
        'handlers': handlers,
        'disable_existing_loggers': False,
        'loggers': {
            'default': {
                'level': 'DEBUG',
            },
        },
        'root': {
            'level': 'DEBUG',
            'handlers': list(handlers.keys()),
        },
    })
    logging.info("Hello world.")


def main():
    setup_logging()
    parser = create_parser()
    args = parser.parse_args()

    with args.input_json.open() as f:
        configuration = json.load(f)

    start = time.time()
    dread_patcher.patch_extracted(
        args.input_path,
        args.output_path,
        configuration,
    )
    end = time.time()
    print("Patcher took {:.03f} seconds".format(end - start))



import io
from typing import Any
from open_dread_rando.patcher_editor import PatcherEditor
from mercury_engine_data_structures.formats.ini import Ini


def apply_cosmetic_patches(editor: PatcherEditor, cosmetic: dict):
    edit_config(editor, cosmetic["config"])


def edit_config(editor: PatcherEditor, config_patches: dict[str, dict[str, Any]]):
    config_file = editor.get_file('config.ini', Ini)

    for section, items in config_patches.items():
        for k, v in items.items():
            config_file.config[section][k] = Ini.parse_option(v)



from mercury_engine_data_structures.formats import Txt

from open_dread_rando.patcher_editor import PatcherEditor

# may want to edit all the localization files?
ALL_TEXT_FILES = {
    # "eu_dutch.txt",
    # "eu_french.txt",
    # "eu_german.txt",
    # "eu_italian.txt",
    # "eu_spanish.txt",
    # "japanese.txt",
    # "korean.txt",
    # "russian.txt",
    # "simplified_chinese.txt",
    # "traditional_chinese.txt",
    "us_english.txt",
    # "us_french.txt",
    # "us_spanish.txt"
}


def patch_text(editor: PatcherEditor, key: str, value: str):
    for text_file in ALL_TEXT_FILES:
        text = editor.get_file(f"system/localization/{text_file}", Txt)
        text.strings[key] = value


def get_text(editor: PatcherEditor, key: str, text_file: str = "us_english.txt") -> str:
    text = editor.get_file(f"system/localization/{text_file}", Txt)
    return text.strings[key]


def apply_text_patches(editor: PatcherEditor, patches: dict[str, str]):
    for k, v in patches.items():
        patch_text(editor, k, v)


def patch_hints(editor: PatcherEditor, hints: list[dict]):
    # patch mission log titles
    for lvl in {"AQUA", "CAVE", "MAGMA", "FOREST", "LAB", "SANC", "SHIP"}:
        mlog_key = f"MLOG_ADAM_{lvl}"
        original = get_text(editor, mlog_key)
        new = original.replace("Adam Briefing", "a hint")
        patch_text(editor, mlog_key, new)

    for hint in hints:
        ap_actor = editor.resolve_actor_reference(hint["accesspoint_actor"])
        usable = ap_actor.pComponents.USABLE

        # don't lock any doors
        usable.vDoorsToChange = []
        usable.wpThermalDevice = ""

        # update hint text
        hint_id = f"DIAG_ADAM_{hint['hint_id']}"
        string_key = f"{hint_id}_PAGE_1"

        usable.tCaptionList = {
            hint_id: [string_key]
        }
        patch_text(editor, string_key, hint["text"])


def patch_credits(editor: PatcherEditor):
    text = editor.get_file("system/localization/credits.txt", Txt)
    ordered_credits = list(text.strings.items())

    rando_credits = [
        ("Randomizer Credits", "_TITLE"),

        ("Game Patching", "_SUBTITLE"),
        ("Henrique 'Darkszero' Gemignani", ""),
        ("duncathan_salt", ""),

        ("Logic Database", "_SUBTITLE"),
        ("KirbymastaH", ""),
        ("Dyceron", ""),
        ("XenoWars", ""),

        ("Additional Art", "_SUBTITLE"),
        ("BigSharksZ", ""),
        ("SkyTheLucario", ""),

        ("     ", "_SUBTITLE"),
        ("With contributions from many others.", ""),
        ("     ", "_SUBTITLE"),
        ("     ", "_SUBTITLE"),
    ]

    ordered_credits[1:1] = [
        (f"CREDIT_R_{i:03}{prefix}", item)
        for i, (item, prefix) in enumerate(rando_credits)
    ]
    text.strings = {k: v for k, v in ordered_credits}



import shutil
from pathlib import Path
from typing import NamedTuple, Tuple

import ips
import keystone

VERSIONS = {
    "1.0.0": "49161D9CCBC15DF944D0B6278A3C446C006B0BE8",
    "2.1.0": "646761F643AFEBB379EDD5E6A5151AF2CEF93DC1",
}
NOP = bytes.fromhex('1F2003D5')


class NSOPatch(ips.Patch):
    def __init__(self):
        # always use IPS32
        super().__init__(True)

    def add_record(self, offset, content, rle_size=-1):
        # NSO files have a 0x100 byte header which is not accounted
        # for in Ghidra/IDA but is accounted for in the IPS
        return super().add_record(offset + 0x100, content, rle_size)


class VersionedPatch(dict):
    def __missing__(self, key):
        return self[max(self.keys())]
    # use the highest version with a patch defined

class AsmVersion(NamedTuple):
    offset: int
    replacements: dict[str, str]


class AsmPatch:
    assembler = keystone.Ks(keystone.KS_ARCH_ARM64, keystone.KS_MODE_LITTLE_ENDIAN)

    def __init__(self, asm_template: str, versions: dict[str, AsmVersion]) -> None:
        self.asm_template = asm_template
        self.versions = versions
    
    def _asm(self, version: str) -> bytes:
        return self.asm_template.format_map(self.versions[version].replacements).encode('ascii')

    def _data(self, version: str) -> bytes:
        encoding, count = self.assembler.asm(self._asm(version), self.versions[version].offset, True)
        return encoding
    
    def patch(self, version: str) -> Tuple[int, bytes]:
        if version not in self.versions:
            return None, None
        return self.versions[version].offset, self._data(version)


def _patch_corpius(patch: ips.Patch, version: str, configuration: dict):
    # patches corpius to not give the phantom cloak, and not to display the
    # "Upgrading suit for Aeion compatibility" message which causes softlocks
    grant_item_none = VersionedPatch({
        "1.0.0": (0x00d94890, bytes.fromhex('A13F00D0 21D81591')),
        "2.1.0": (0x00da7380, bytes.fromhex('81410090 21B82091')),
    })
    stub_aeion_message = VersionedPatch({
        "1.0.0": (0x011a1f4c, NOP),
        "2.1.0": (0x011e7f1c, NOP),
    })

    for p in [grant_item_none, stub_aeion_message]:
        offset, data = p[version]
        patch.add_record(offset, data)


def _add_version_sentinel(patch: ips.Patch, version: str):
    # Replace `IsInSTEAM_PC_FINAL_RETAIL` with a custom name, so lua code can know if the exefs was properly patched
    randomized_sentinel = {
        "1.0.0": (0x0158c3c2, b"HasRandomizerPatches\x00"),
        "2.1.0": (0x015d9780, b"HasRandomizerPatches\x00"),
    }

    offset, data = randomized_sentinel[version]
    patch.add_record(offset, data)


debug_input = AsmPatch(
    Path(__file__).parent.joinpath("files/exefs_patches/debug_input.s").read_text(),
    {
        "1.0.0": AsmVersion(0x010525f0, {
            "getNpadState1": "0x011f3630",
            "getNpadState2": "0x011f3640",
            "lua_pushinteger": "0x01060730",
        }),
        "2.1.0": AsmVersion(0x010943a0, {
            "getNpadState1": "0x012393d0",
            "getNpadState2": "0x012393e0",
            "lua_pushinteger": "0x010a25f0",
        }),
    }
)

def _add_debug_input(patch: ips.Patch, version: str):
    offset, data = debug_input.patch(version)
    if offset is not None:
        patch.add_record(offset, data)


def patch_exefs(exefs_patches: Path, configuration: dict):
    shutil.rmtree(exefs_patches, ignore_errors=True)
    exefs_patches.mkdir(parents=True, exist_ok=True)

    for version, exefs_hash in VERSIONS.items():
        patch = NSOPatch()
        _patch_corpius(patch, version, configuration)
        _add_version_sentinel(patch, version)
        _add_debug_input(patch, version)
        exefs_patches.joinpath(f"{exefs_hash}.ips").write_bytes(bytes(patch))


def include_depackager(exefs_path: Path):
    shutil.rmtree(exefs_path, ignore_errors=True)
    exefs_path.mkdir(parents=True, exist_ok=True)

    dread_depackager = Path(__file__).parent.joinpath("files", "dread_depackager")
    shutil.copytree(dread_depackager, exefs_path, dirs_exist_ok=True)



from unittest.mock import MagicMock

from open_dread_rando import lua_util


def test_create_script_copy_exists():
    editor = MagicMock()
    editor.does_asset_exists.return_value = True

    # Run
    lua_util.create_script_copy(editor, "system/scripts/init")

    # Assert
    editor.does_asset_exists.assert_called_once_with("system/scripts/init_original.lc")
    editor.add_new_asset.assert_not_called()



from open_dread_rando.patcher_editor import PatcherEditor


def is_door(actor):
    return "LIFE" in actor.pComponents and "CDoorLifeComponent" == actor.pComponents.LIFE["@type"]


def reference_for_link(link: str, scenario: str) -> dict:
    split_link = link.split(":")
    if len(split_link) != 7:
        raise ValueError(f"Expected 7 components in {link}, got {len(split_link)}")

    layer = split_link[4]
    actor = split_link[6]

    return {
        "scenario": scenario,
        "layer": layer,
        "actor": actor,
    }


_ACTOR_DEF_FOR_TYPE = {
    "frame": "actors/props/doorframe/charclasses/doorframe.bmsad",
    "power_beam": "actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad",
    "charge_beam": "actors/props/doorchargecharge/charclasses/doorchargecharge.bmsad",
    "wide_beam": "actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad",
    "plasma_beam": "actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad",
    "wave_beam": "actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad",
    "missile": "actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad",
    "super_missile": "actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad",
    "grapple_beam": "actors/props/doorgrapplegrapple/charclasses/doorgrapplegrapple.bmsad",
    "phantom_cloak": "actors/props/doorpresencepresence/charclasses/doorpresencepresence.bmsad",
    "phase_shift": "actors/props/doorshutter/charclasses/doorshutter.bmsad",
}


def patch_door(editor: PatcherEditor, door: dict):
    actor = editor.resolve_actor_reference(door["actor"])

    if not is_door(actor):
        raise ValueError(f"Actor at {door['actor']} is not a door.")

    door_type = door["door_type"]
    need_shield = door_type in {"wide_beam", "plasma_beam", "wave_beam", "missile", "super_missile"}

    life_comp = actor.pComponents.LIFE
    if need_shield:
        raise NotImplementedError("Adding shields to doors is not implemented")
    else:
        for link_name in ["wpLeftDoorShieldEntity", "wpRightDoorShieldEntity"]:
            link = life_comp[link_name]
            if link != '{EMPTY}':
                editor.remove_entity(reference_for_link(link, door["actor"]["scenario"]), "mapBlockages")
                life_comp[link_name] = '{EMPTY}'

        actor.oActorDefLink = f"actordef:{_ACTOR_DEF_FOR_TYPE[door_type]}"
        for pkg_name in editor.get_level_pkgs(door["actor"]["scenario"]):
            editor.ensure_present(pkg_name, _ACTOR_DEF_FOR_TYPE[door_type])
